"""Contains utilities for training machine learning and deep learning models.

This includes prebuilt models for common ARPES tasks, and tools which facilitate integration
with other libraries including pytorch.
"""
